;; Generated package description from project.el  -*- no-byte-compile: t -*-
(define-package "project" "0.11.1.0.20260102.5922" "Operations on the current project" '((emacs "26.1") (xref "1.7.0")) :commit "2175e9a192d436c94ab88fef8858361197b297bb" :url "https://elpa.gnu.org/packages/project.html")
